import collections
import logging
import pathlib
import typing

from odd.artifact import Addon
from odd.const import MANIFEST_FILENAMES

_LOG = logging.getLogger(__name__)


def find_manifest(path: pathlib.Path) -> typing.Optional[pathlib.Path]:
    for child in path.iterdir():
        if child.is_file() and child.name in MANIFEST_FILENAMES:
            return child
    return None


def discover_addons(
    dir_path: pathlib.Path,
    odoo_version: int,
    *,
    allow_duplicates: bool = False,
    allow_not_installable: bool = True,
) -> typing.Generator[Addon, None, None]:
    """
    :param allow_duplicates: If False, addons with the same name are ignored.
    :param allow_not_installable: If False, addons that are marked as not
        installable are ignored.
    """
    dirs = collections.deque((dir_path,))
    seen: typing.Set[str] = set()

    def check_dir(
        path: pathlib.Path, descend: bool = True
    ) -> typing.Generator[Addon, None, None]:
        manifest_path = find_manifest(path)
        if manifest_path:
            addon_name = manifest_path.parent.name
            if allow_duplicates or addon_name not in seen:
                addon = Addon.from_manifest_path(manifest_path, odoo_version)
                if allow_not_installable or addon.manifest.get("installable", True):
                    yield addon
            if not allow_duplicates:
                if addon_name in seen:
                    _LOG.warning(
                        'Addon with the same name "%s" found at path: %s, '
                        "it will be ignored",
                        addon_name,
                        path,
                    )
                else:
                    seen.add(addon_name)
        else:
            if descend:
                dirs.append(path)

    yield from check_dir(dir_path, descend=False)
    while dirs:
        dir_ = dirs.popleft()
        for child in dir_.iterdir():
            if child.is_dir():
                yield from check_dir(child)
