import ast
import dataclasses
import functools
import typing

import parso

from odd.const import UNKNOWN
from odd.artifact.issue import Position

STRING_NODE_TYPES = frozenset(("string", "strings"))
ATOM_EXPR_NODE_TYPES = frozenset(("atom_expr", "power"))  # "power" is PY2.


@functools.lru_cache(maxsize=16)
def get_grammar(odoo_version: int) -> parso.Grammar:
    if odoo_version >= 13:
        return parso.load_grammar(version="3.6")
    return parso.load_grammar(version="3.5")


@dataclasses.dataclass
class Call:
    names: typing.Tuple[str, ...]
    args: typing.Tuple[typing.Any, ...]
    kwargs: typing.Dict[str, typing.Any]
    start_pos: Position
    end_pos: Position


def iter_calls(main_node: parso.tree.NodeOrLeaf) -> typing.Generator[Call, None, None]:
    for node in iter_node_type(main_node, *ATOM_EXPR_NODE_TYPES):
        yield from _parse_call(node)


def is_op(node: parso.tree.NodeOrLeaf, *values: str) -> bool:
    return node.type == "operator" and node.value in values


def parse_call_args(
    arg_nodes: typing.Iterable[parso.tree.NodeOrLeaf],
) -> typing.Tuple[typing.Tuple[typing.Any, ...], typing.Dict[str, typing.Any]]:
    args: typing.List[typing.Any] = []
    kwargs = {}

    def parse_argument(arg_node):
        if is_string_node(arg_node):
            args.append(get_string_node_value(arg_node))
        elif arg_node.type in ("number", "keyword"):
            args.append(get_node_value(arg_node))
        elif arg_node.type == "name":
            args.append(UNKNOWN)
        elif is_op(arg_node, ","):
            ...
        elif (
            isinstance(arg_node, parso.tree.Node)
            and len(arg_node.children) == 2
            and is_op(arg_node.children[0], "*", "**")
            and arg_node.children[1].type == "name"
        ):
            # *a, **kw
            ...
        elif (
            isinstance(arg_node, parso.tree.Node)
            and len(arg_node.children) == 3
            and arg_node.children[0].type == "name"
            and is_op(arg_node.children[1], "=")
        ):
            name, op, value_node = arg_node.children
            kwargs[name.value] = get_node_value(value_node)
        else:
            # print(f"Unexpected args: {arg_node!r}")
            args.append(UNKNOWN)

    for arg_node in arg_nodes:
        if arg_node.type == "arglist":
            for child_arg in arg_node.children:
                parse_argument(child_arg)
        else:
            parse_argument(arg_node)

    return tuple(args), kwargs


def _parse_call(node: parso.tree.Node) -> typing.Generator[Call, None, None]:
    name_parts = []
    for child in node.children:
        if child.type == "name":
            name_parts.append(child.value)
        elif child.type == "trailer":
            if not child.children:
                continue

            if is_parenthesized(child):
                args, kwargs = parse_call_args(child.children[1:-1])
                yield Call(
                    tuple(name_parts), args, kwargs, child.start_pos, child.end_pos
                )
                continue

            for sub_child in child.children:
                if is_op(sub_child, "."):
                    ...
                elif sub_child.type == "name":
                    name_parts.append(sub_child.value)
                else:
                    ...
                    # print(f"Unhandled subchild: {sub_child!r}")
        else:
            ...
            # print(f"Unexpected node while parsing call: {child!r}")


def walk(
    node: parso.tree.NodeOrLeaf
) -> typing.Generator[parso.tree.NodeOrLeaf, None, None]:
    yield node
    try:
        children = node.children
    except AttributeError:
        pass
    else:
        for child in children:
            yield from walk(child)


def iter_node_type(
    main_node: parso.tree.NodeOrLeaf, *types: str
) -> typing.Generator[parso.tree.NodeOrLeaf, None, None]:
    for node in walk(main_node):
        if node.type in types:
            yield node


def iter_children_node_type(
    node: parso.tree.Node, *types: str
) -> typing.Generator[parso.tree.NodeOrLeaf, None, None]:
    for child in node.children:
        if child.type in types:
            yield child


def get_node_value(node: parso.tree.NodeOrLeaf) -> typing.Any:
    if is_string_node(node):
        return get_string_node_value(node)
    elif node.type == "keyword" and node.value in ("True", "False"):
        return node.value == "True"
    elif node.type == "keyword" and node.value == "None":
        return None
    elif node.type in ("keyword", "number"):
        return ast.literal_eval(node.value)
    elif node.type == "name":
        return UNKNOWN
    else:
        try:
            return ast.literal_eval(node.get_code().strip())
        except (SyntaxError, ValueError, IndentationError):
            return UNKNOWN


def get_string_node_value(node: parso.tree.NodeOrLeaf) -> str:
    if node.type in ("strings", "arith_expr", "atom"):
        if node.type == "arith_expr":
            children = node.children[::2]
        elif node.type == "atom":
            children = node.children[1:-1]
        else:
            children = node.children
        return "".join(get_string_node_value(child) for child in children)
    elif node.type == "string":
        return node._get_payload()
    else:
        raise TypeError(f"Unexpected node type: {node.type}")


def is_string_arith_expr(node: parso.tree.NodeOrLeaf) -> bool:
    """Returns True if `node` is `"a" + [... +] "z"`."""
    if node.type != "arith_expr":
        return False

    if len(node.children) % 2 != 1:
        return False

    for string_node in node.children[::2]:
        if string_node.type not in STRING_NODE_TYPES:
            return False

    for op_node in node.children[1::2]:
        if not is_op(op_node, "+"):
            return False

    return True


def is_string_atom(node: parso.tree.NodeOrLeaf) -> bool:
    """Returns True if `node` is `("a" [...] "z")`."""
    if node.type != "atom":
        return False

    if not is_parenthesized(node):
        return False

    for child in node.children[1:-1]:
        if child.type not in STRING_NODE_TYPES:
            return False

    return True


def is_string_node(node: parso.tree.NodeOrLeaf) -> bool:
    return (
        node.type in STRING_NODE_TYPES
        or is_string_arith_expr(node)
        or is_string_atom(node)
    )


def is_parenthesized(node: parso.tree.Node) -> bool:
    if len(node.children) < 2:
        return False
    head, tail = node.children[0], node.children[-1]
    return is_op(head, "(") and is_op(tail, ")")


def get_name(node: parso.tree.NodeOrLeaf) -> typing.Generator[str, None, None]:
    if node.type == "name":
        yield node.value
    elif node.type in ("atom_expr", "trailer"):
        for child in node.children:
            yield from get_name(child)
    elif is_op(node, "."):
        pass
    else:
        raise ValueError(f"Unexpected node type: {child.type}")


def get_bases(
    node: typing.Optional[parso.tree.NodeOrLeaf]
) -> typing.List[typing.Tuple[str, ...]]:
    if node is None:
        return []
    elif node.type == "arglist" or node.type == "testlist":
        return [
            tuple(get_name(c))
            for c in node.children
            if c.type == "name" or c.type in ATOM_EXPR_NODE_TYPES
        ]
    else:
        return [tuple(get_name(node))]


def consume_name(
    node: parso.tree.Node
) -> typing.Tuple[typing.List[str], typing.List[parso.tree.NodeOrLeaf]]:
    name_parts, leftover_nodes = [], []
    for child in node.children:
        if child.type == "name":
            name_parts.append(child.value)
        elif (
            child.type == "trailer"
            and len(child.children) == 2
            and is_op(child.children[0], ".")
            and child.children[1].type == "name"
        ):
            name_parts.append(child.children[1].value)
        else:
            leftover_nodes.append(child)
    return name_parts, leftover_nodes


def get_model_name(params) -> typing.Optional[str]:
    _name = params.get("_name")
    if _name:
        return _name
    _inherit = params.get("_inherit")
    if _inherit:
        if isinstance(_inherit, list):
            # `_inherit = []`
            if len(_inherit) == 0:
                return None
            # `_inherit = ['foo']`
            elif len(_inherit) == 1:
                return _inherit[0]
            # `_inherit = ['foo', 'bar', ...]`, currently not supported.
            else:
                raise ValueError(
                    f"Unexpected number of `_inherit` models: {len(_inherit)}"
                )
        else:
            return _inherit
    return None
