import typing

import parso

from odd.artifact import (
    AddonPath,
    FieldArg,
    FieldDefinition,
    FieldKwarg,
    ModelDefinition,
    PythonModule,
)
from odd.ast_utils import (
    ATOM_EXPR_NODE_TYPES,
    STRING_NODE_TYPES,
    consume_name,
    get_bases,
    get_grammar,
    get_model_name,
    get_node_value,
    get_string_node_value,
    iter_children_node_type,
)
from odd.const import SUPPORTED_ODOO_VERSIONS, UNKNOWN
from odd.plugin import Plugin
from odd.utils import expand_version_list

FIELD_TYPE_VERSION_MAP = expand_version_list(
    {
        ">=8": {
            "Binary",
            "Boolean",
            "Char",
            "Date",
            "Datetime",
            "Float",
            "Html",
            "Id",
            "Integer",
            "Many2many",
            "Many2one",
            "One2many",
            "Reference",
            "Selection",
            "Serialized",
            "Text",
        },
        ">=9": {"Monetary"},
        ">=13": {"Image", "Many2oneReference"},
    },
    *SUPPORTED_ODOO_VERSIONS,
    result_cls=set,
)


def _get_model_params(
    classdef_node: parso.python.tree.Class
) -> typing.Generator[typing.Tuple[str, typing.Any], None, None]:
    def handle_expr_stmt(
        node
    ) -> typing.Generator[typing.Tuple[str, typing.Any], None, None]:
        value = get_node_value(node.get_rhs())
        for name_node in node.get_defined_names():
            name = name_node.value
            if name.startswith("_"):
                yield name, value

    for node in iter_children_node_type(
        classdef_node.get_suite(), "simple_stmt", "expr_stmt"
    ):
        if node.type == "simple_stmt":
            for expr_stmt in iter_children_node_type(node, "expr_stmt"):
                yield from handle_expr_stmt(expr_stmt)
        else:
            yield from handle_expr_stmt(node)


class PythonEmitter(Plugin):
    _handles = {"addon_path"}
    _emits = {"python_module"}

    def on_addon_path(self, addon_path: AddonPath):
        if addon_path.path.suffix != ".py":
            return
        with addon_path.path.open(mode="rb") as f:
            node = get_grammar(addon_path.addon.odoo_version).parse(f.read())
        yield PythonModule(addon_path.addon, addon_path.path, node)


class ModelDefinitionEmitter(Plugin):
    _handles = {"python_module"}
    _emits = {"model_definition"}

    def on_python_module(self, python_module: PythonModule):
        for classdef in python_module.node.iter_classdefs():
            model_params = dict(_get_model_params(classdef))

            if not get_model_name(model_params):
                continue

            yield ModelDefinition(
                addon=python_module.addon,
                path=python_module.path,
                node=classdef,
                class_name=classdef.name.value,
                params=model_params,
                bases=get_bases(classdef.get_super_arglist()),
            )


class FieldDefinitionEmitter(Plugin):
    _handles = {"model_definition"}
    _emits = {"field_definition"}

    def on_model_definition(self, model: ModelDefinition):
        field_types = FIELD_TYPE_VERSION_MAP[model.addon.odoo_version]

        def handle_expr_stmt(node):
            names = [n.value for n in node.get_defined_names()]
            field_node = node.get_rhs()

            if field_node.type not in ATOM_EXPR_NODE_TYPES:
                return

            name_parts, leftover_nodes = consume_name(field_node)
            if not name_parts or len(name_parts) > 3 or len(leftover_nodes) != 1:
                return

            if len(name_parts) == 3 and (
                name_parts[0] not in ("odoo", "openerp") or name_parts[1] != "fields"
            ):
                return
            elif len(name_parts) == 2 and name_parts[0] != "fields":
                return

            field_class = name_parts[-1]

            # E.g. `foo = Someclass()``. Since we have no extra ways to
            # check if `Someclass` is a field, skip this case.
            if len(name_parts) == 1 and field_class not in field_types:
                return

            args, kwargs = [], []
            arg_node_list = []
            for child in leftover_nodes[0].children:
                if child.type == "operator":
                    continue
                elif child.type == "argument":
                    arg_node_list = [child]
                    break
                elif child.type == "arglist":
                    arg_node_list = [
                        arg for arg in child.children if arg.type == "argument"
                    ]
                    break
                else:
                    args.append(
                        FieldArg(get_node_value(child), child.start_pos, child.end_pos)
                    )

            for arg_node in arg_node_list:
                if arg_node.children[0].type == "name":
                    kwarg_name = arg_node.children[0].value

                    if arg_node.children[2].type in STRING_NODE_TYPES:
                        kwarg_value = get_string_node_value(arg_node.children[2])
                    else:
                        kwarg_value = UNKNOWN

                    kwargs.append(
                        FieldKwarg(
                            name=kwarg_name,
                            value=kwarg_value,
                            start_pos=arg_node.start_pos,
                            end_pos=arg_node.end_pos,
                        )
                    )

            for name in names:
                yield FieldDefinition(
                    name=name,
                    class_name=field_class,
                    model=model,
                    start_pos=field_node.start_pos,
                    end_pos=field_node.end_pos,
                    args=args,
                    kwargs=kwargs,
                )

        for node in iter_children_node_type(
            model.node.get_suite(), "simple_stmt", "expr_stmt"
        ):
            if node.type == "simple_stmt":
                for expr_stmt in iter_children_node_type(node, "expr_stmt"):
                    yield from handle_expr_stmt(expr_stmt)
            else:
                yield from handle_expr_stmt(node)
